from __future__ import annotations

from typing import Any, Dict, List, Iterable, Optional, Sequence
from itertools import cycle
from threading import Lock
from collections import defaultdict

from .types import IsaacusSageMakerRuntimeEndpoint


class _RoundRobin:
    def __init__(self, items: Iterable[Any]) -> None:
        self._items = list(items)
        self._it = cycle(self._items) if self._items else None

    def next(self) -> Optional[Any]:
        if self._it is None:
            return None

        return next(self._it)


class Router:
    def __init__(self, endpoints: Sequence[IsaacusSageMakerRuntimeEndpoint]) -> None:
        assert endpoints, "At least one endpoint must be provided."
        assert all(isinstance(e, IsaacusSageMakerRuntimeEndpoint) for e in endpoints), (
            "All provided endpoints must be of the type `IsaacusSageMakerRuntimeEndpoint`."
        )
        assert all(e.name for e in endpoints), "All provided endpoints must have a non-empty name."

        self._lock = Lock()
        self._all_rr = _RoundRobin([e for e in endpoints if e.models is None])

        by_model: Dict[str, List[IsaacusSageMakerRuntimeEndpoint]] = defaultdict(list)

        for e in endpoints:
            if e.models is not None:
                for model in e.models:
                    by_model[model].append(e)

        for e in endpoints:
            if e.models is None:
                for model in by_model.keys():
                    by_model[model].append(e)

        self._by_model_rr: Dict[str, _RoundRobin] = {
            model: _RoundRobin(endpoints) for model, endpoints in by_model.items()
        }

    def pick(self, model: Optional[str]) -> Optional[IsaacusSageMakerRuntimeEndpoint]:
        with self._lock:
            if model and model in self._by_model_rr:
                n = self._by_model_rr[model].next()

                if n is not None:
                    return n

            return self._all_rr.next()



import os
import sys
import types
import pathlib
from typing import Any
import pytest

# Ensure `src` (editable dev layout) is importable when running from the repo root
try:
    import isaacus_sagemaker  # noqa: F401
except Exception:  # pragma: no cover - only runs in local dev
    repo_root = pathlib.Path(__file__).resolve().parents[1]
    src = repo_root / "src"
    if src.exists():
        sys.path.insert(0, str(src))

# Lightweight fakes for AWS signing and boto3 session ------------------------------------

class _DummyAWSRequest:
    """
    Drop-in-enough replacement for botocore.awsrequest.AWSRequest used in tests.
    Only the attributes accessed by the package code are implemented.
    """
    def __init__(self, *, method: str, url: str, data: bytes, headers: dict[str, str]):
        self.method = method
        self.url = url
        # start with a copy so the signer can mutate
        self.headers = dict(headers)
        self.data = data


class _DummySigV4Auth:
    """
    Very small stand-in for botocore.auth.SigV4Auth that simply stamps a header.
    """
    def __init__(self, credentials: Any, service_name: str, region_name: str) -> None:
        # Record what we were given for assertions if needed
        self.credentials = credentials
        self.service_name = service_name
        self.region_name = region_name

    # mimic the API of botocore.auth.SigV4Auth
    def add_auth(self, aws_request: _DummyAWSRequest) -> None:
        # Stamp an Authorization header so tests can assert it was "signed"
        aws_request.headers["Authorization"] = f"AWS4-HMAC-SHA256 Dummy region={self.region_name} service={self.service_name}"


class _DummyCreds:
    def __init__(self, access_key="AKID", secret_key="SECRET", token="TOKEN"):
        self.access_key = access_key
        self.secret_key = secret_key
        self.token = token


class _DummyBoto3Session:
    # class-level counter to let tests assert we only build a session once per profile
    constructed = 0

    def __init__(self, **kwargs: Any) -> None:
        _DummyBoto3Session.constructed += 1
        # allow explicit region override; otherwise default
        self.region_name = kwargs.get("region_name") or "us-east-1"
        self.profile_name = kwargs.get("profile_name")

    def get_credentials(self):
        return _DummyCreds()


@pytest.fixture
def fake_aws(monkeypatch):
    """
    Monkeypatch AWSRequest, SigV4Auth and boto3.Session to predictable dummies.
    """
    import isaacus_sagemaker.runtime_client as rc

    monkeypatch.setattr(rc, "AWSRequest", _DummyAWSRequest, raising=True)
    monkeypatch.setattr(rc, "SigV4Auth", _DummySigV4Auth, raising=True)

    # Patch boto3.Session used by the runtime client
    import boto3
    monkeypatch.setattr(boto3, "Session", _DummyBoto3Session, raising=True)

    # Reset counter before each test that uses this fixture
    _DummyBoto3Session.constructed = 0

    return {
        "AWSRequest": _DummyAWSRequest,
        "SigV4Auth": _DummySigV4Auth,
        "Boto3Session": _DummyBoto3Session,
    }


@pytest.fixture
def endpoints():
    from isaacus_sagemaker.types import IsaacusSageMakerRuntimeEndpoint

    return [
        # default endpoint (serves all models)
        IsaacusSageMakerRuntimeEndpoint(name="default-endpoint", region="ap-southeast-2"),
        # model-specific endpoint
        IsaacusSageMakerRuntimeEndpoint(name="embed-endpoint", region="ap-southeast-2", models=["kanon-2-embedder"]),
    ]




from __future__ import annotations

from typing import Any, Dict, Literal, Optional, Sequence

import msgspec

HTTP_METHODS = Literal["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE", "PATCH"]


class IsaacusSageMakerInvocationRequest(msgspec.Struct, frozen=True):
    path: str
    """The Isaacus API path (including any version prefix) to proxy the request to (e.g., `"/v1/embeddings"`)."""

    method: HTTP_METHODS = "POST"
    """The HTTP method to use for the proxied request (e.g., `"POST"`). Defaults to `"POST"`."""

    headers: Optional[Dict[str, str]] = None
    """Optional HTTP headers to include in the proxied request. Defaults to `None` in which case no extra headers are added."""

    data: Optional[Any] = None
    """The request body to include in the proxied request. Defaults to `None` in which case no request body is sent."""


class IsaacusSageMakerRuntimeEndpoint(msgspec.Struct, frozen=True):
    name: str
    """The name of the SageMaker endpoint."""

    region: Optional[str] = None
    """The AWS region where the SageMaker endpoint is deployed. This overrides any region specified at the client-level."""

    profile: Optional[str] = None
    """The AWS profile to use when accessing the SageMaker endpoint. This overrides any profile specified at the client-level."""

    models: Optional[Sequence[str]] = None
    """The IDs of models served by this endpoint. If `None`, it is assumed the endpoint serves all models."""



from isaacus_sagemaker.types import IsaacusSageMakerInvocationRequest, IsaacusSageMakerRuntimeEndpoint

def test_struct_defaults_and_frozen():
    # frozen=True -> attributes are immutable
    req = IsaacusSageMakerInvocationRequest(path="/v1/embeddings")
    assert req.method == "POST"
    assert req.headers is None and req.data is None
    try:
        req.path = "/new"  # type: ignore[attr-defined]
        assert False, "Should be frozen/immutable"
    except Exception:
        pass

    ep = IsaacusSageMakerRuntimeEndpoint(name="abc")
    assert ep.region is None and ep.profile is None and ep.models is None


from __future__ import annotations

import os

from typing import Any, Dict, Union, Mapping, Optional, Sequence

import boto3
import httpx
import orjson
import msgspec
import botocore.credentials

from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest

from .types import IsaacusSageMakerRuntimeEndpoint, IsaacusSageMakerInvocationRequest
from ._router import Router

_UNPROXIABLE_HEADERS = {
    "host",
    "content-length",
    "connection",
    "keep-alive",
    "proxy-authenticate",
    "proxy-authorization",
    "te",
    "trailers",
    "transfer-encoding",
    "upgrade",
    "authorization",
}


def _extract_body(message: Union[httpx.Request, httpx.Response]) -> Any:
    content_type = message.headers.get("content-type", "")

    if "application/json" in content_type:
        body = orjson.loads(message.content)

    else:
        body = message.content

        if body:
            raise ValueError("Only message bodies with the content type `application/json` are supported.")

    return body


def _get_model(body: Any) -> Optional[str]:
    if isinstance(body, dict) and "model" in body:
        return body["model"]

    return None


def _strip_unproxiable_headers(headers: Mapping[str, str]) -> Dict[str, str]:
    return {k: v for k, v in headers.items() if k.lower() not in _UNPROXIABLE_HEADERS}


def _make_payload(
    req: httpx.Request,
    body: Any,
) -> bytes:
    invocation = IsaacusSageMakerInvocationRequest(
        path=req.url.path,
        method=req.method,
        headers=_strip_unproxiable_headers(req.headers),
        data=body,
    )
    payload = msgspec.json.encode(invocation)

    return payload


def _translate_response(resp: httpx.Response) -> httpx.Response:
    # Pass through non-error responses as-is.
    if resp.status_code < 400:
        return resp

    # Attempt to parse the response body as JSON otherwise return the original response.
    content_type = resp.headers.get("content-type", "")
    payload: Optional[Dict] = None

    if "application/json" in content_type:
        payload = orjson.loads(resp.content)

    if not isinstance(payload, dict):
        return resp

    # Attempt to extract the original status code and message from the payload.
    original_status = payload.get("OriginalStatusCode")

    if not original_status:
        return resp

    original_message = payload.get("OriginalMessage")

    if not original_message:
        return resp

    # Attempt to parse the original message as JSON.
    try:
        original_message = orjson.loads(original_message)

    except Exception:
        return resp

    # Build the translated response.
    headers = resp.headers.copy()
    headers["content-type"] = "application/json"

    return httpx.Response(
        status_code=int(original_status),
        headers=headers,
        json=original_message,
        request=resp.request,
    )


def _get_creds(
    endpoint: IsaacusSageMakerRuntimeEndpoint,
    creds: Dict[Optional[str], botocore.credentials.Credentials],
    default_profile: Optional[str] = None,
    session_kwargs: Optional[Dict[str, Any]] = None,
) -> botocore.credentials.Credentials:
    profile = endpoint.profile or default_profile

    if profile not in creds:
        session_kwargs = dict(session_kwargs or {})
        
        if profile is not None:
            session_kwargs["profile_name"] = profile

        session = boto3.Session(**session_kwargs)
        endpoint_creds = session.get_credentials()

        if endpoint_creds is None:
            raise RuntimeError(f"Could not retrieve AWS credentials for profile '{profile}'.")

        creds[profile] = endpoint_creds

    return creds[profile]


def _build_request(
    self: IsaacusSageMakerRuntimeHTTPClient | AsyncIsaacusSageMakerRuntimeHTTPClient,
    req: httpx.Request,
) -> httpx.Request:
    # Extract the request body.
    req_body = _extract_body(req)

    # Determine the model to route the request to if one was specified.
    model = _get_model(req_body)

    # Pick an endpoint to route the request to.
    endpoint = self._router.pick(model)

    if endpoint is None:
        raise RuntimeError(f"No SageMaker endpoints registered for model '{model}'.")

    # Construct the SageMaker invocation payload.
    req_payload = _make_payload(req, req_body)

    # Build the request headers.
    req_headers = {
        "Content-Type": "application/json",
        "Accept": req.headers.get("accept", "application/json"),
    }

    # Construct the URL for the request.
    region = endpoint.region or self._region
    assert region is not None, (
        "AWS region must be specified either at the endpoint-level, client-level or environment-level."
    )
    url = f"https://runtime.sagemaker.{region}.amazonaws.com/endpoints/{endpoint.name}/invocations"

    # Get AWS credentials for the endpoint.
    creds = _get_creds(endpoint, self._creds, default_profile=self._profile, session_kwargs=self._boto_session_kwargs)

    # Sign the request payload with AWS SigV4.
    req_for_sigv4 = AWSRequest(method="POST", url=url, data=req_payload, headers=req_headers)
    SigV4Auth(credentials=creds, service_name="sagemaker", region_name=region).add_auth(req_for_sigv4)
    signed_req_headers = {k: str(v) for k, v in req_for_sigv4.headers.items()}

    # Build the request.
    req = self.build_request(
        method="POST",
        url=url,
        headers=signed_req_headers,
        data=req_payload,
    )

    return req


class IsaacusSageMakerRuntimeHTTPClient(httpx.Client):
    """
    A synchronous Isaacus SDK-compatible HTTP client that proxies requests to SageMaker-deployed Isaacus models through the SageMaker Runtime InvokeEndpoint (`/invocations`) API.

    This client extends `httpx.AsyncClient`.

    Arguments:
        `endpoints` (`Sequence[IsaacusSageMakerRuntimeEndpoint]`): A sequence of SageMaker endpoints to route requests to.
        `region` (`str`, optional): The AWS region where the SageMaker endpoints are deployed. Overriden by any region specified at the endpoint-level. Defaults to `None`, in which case the AWS SDK's default region resolution is used.
        `profile` (`str`, optional): The AWS profile to use when accessing the SageMaker endpoints. Overriden by any profile specified at the endpoint-level. Defaults to `None`, in which case the AWS SDK's default profile resolution is used.
        `boto_session_kwargs` (`Dict[str, Any]`, optional): Additional keyword arguments to pass to the `boto3.Session` constructor when creating the AWS session. Defaults to `None`.
        `**httpx_kwargs` (`Any`): Additional keyword arguments to pass to the `httpx.Client` constructor.
    """

    def __init__(
        self,
        *,
        endpoints: Sequence[IsaacusSageMakerRuntimeEndpoint],
        region: Optional[str] = None,
        profile: Optional[str] = None,
        boto_session_kwargs: Optional[Dict[str, Any]] = None,
        **httpx_kwargs: Any,
    ) -> None:
        super().__init__(**httpx_kwargs)

        self._router = Router(endpoints)

        self._boto_session_kwargs: Dict[str, Any] = boto_session_kwargs or {}
        self._region = (
            region
            or boto3.Session(**self._boto_session_kwargs | ({"profile_name": profile} if profile else {})).region_name
        )
        self._profile = profile
        self._creds: Dict[Optional[str], botocore.credentials.Credentials] = {}

        # If ISAACUS_API_KEY has not been set, set it to `ISAACUS_SAGEMAKER_DOES_NOT_NEED_AN_API_KEY` to avoid errors.
        if "ISAACUS_API_KEY" not in os.environ:
            os.environ["ISAACUS_API_KEY"] = "ISAACUS_SAGEMAKER_DOES_NOT_NEED_AN_API_KEY"

    def send(
        self,
        req: httpx.Request,
        *args: Any,
        **kwargs: Any,
    ) -> httpx.Response:
        # Skip proxying requests to `examples.isaacus.com` (to ensure SDK examples still work as expected).
        if req.url.host.endswith("examples.isaacus.com"):
            return super().send(req, *args, **kwargs)

        # Build the request.
        req = _build_request(self, req)

        # Send the request.
        resp = super().send(req, *args, **kwargs)

        # Translate the response.
        resp = _translate_response(resp)

        return resp


class AsyncIsaacusSageMakerRuntimeHTTPClient(httpx.AsyncClient):
    """
    An asynchronous Isaacus SDK-compatible HTTP client that proxies requests to SageMaker-deployed Isaacus models through the SageMaker Runtime InvokeEndpoint (`/invocations`) API.

    This client extends `httpx.Client`.

    Arguments:
        `endpoints` (`Sequence[IsaacusSageMakerRuntimeEndpoint]`): A sequence of SageMaker endpoints to route requests to.
        `region` (`str`, optional): The AWS region where the SageMaker endpoints are deployed. Overriden by any region specified at the endpoint-level. Defaults to `None`, in which case the AWS SDK's default region resolution is used.
        `profile` (`str`, optional): The AWS profile to use when accessing the SageMaker endpoints. Overriden by any profile specified at the endpoint-level. Defaults to `None`, in which case the AWS SDK's default profile resolution is used.
        `boto_session_kwargs` (`Dict[str, Any]`, optional): Additional keyword arguments to pass to the `boto3.Session` constructor when creating the AWS session. Defaults to `None`.
        `**httpx_kwargs` (`Any`): Additional keyword arguments to pass to the `httpx.Client` constructor.
    """

    def __init__(
        self,
        *,
        endpoints: Sequence[IsaacusSageMakerRuntimeEndpoint],
        region: Optional[str] = None,
        profile: Optional[str] = None,
        boto_session_kwargs: Optional[Dict[str, Any]] = None,
        **httpx_kwargs: Any,
    ) -> None:
        super().__init__(**httpx_kwargs)

        self._router = Router(endpoints)

        self._boto_session_kwargs: Dict[str, Any] = boto_session_kwargs or {}
        self._region = (
            region
            or boto3.Session(**self._boto_session_kwargs | ({"profile_name": profile} if profile else {})).region_name
        )
        self._profile = profile
        self._creds: Dict[Optional[str], botocore.credentials.Credentials] = {}

        # If ISAACUS_API_KEY has not been set, set it to `ISAACUS_SAGEMAKER_DOES_NOT_NEED_AN_API_KEY` to avoid errors.
        if "ISAACUS_API_KEY" not in os.environ:
            os.environ["ISAACUS_API_KEY"] = "ISAACUS_SAGEMAKER_DOES_NOT_NEED_AN_API_KEY"

    async def send(
        self,
        req: httpx.Request,
        *args: Any,
        **kwargs: Any,
    ) -> httpx.Response:
        # Skip proxying requests to `examples.isaacus.com` (to ensure SDK examples still work as expected).
        if req.url.host.endswith("examples.isaacus.com"):
            return await super().send(req, *args, **kwargs)

        # Build the request.
        req = _build_request(self, req)

        # Send the request.
        resp = await super().send(req, *args, **kwargs)

        # Translate the response.
        resp = _translate_response(resp)

        return resp



from typing import IO, Union, Optional, Protocol, TypedDict

from botocore.response import StreamingBody


class InvokeEndpointOutput(TypedDict, total=False):
    Body: StreamingBody
    ContentType: str
    InvokedProductionVariant: str
    CustomAttributes: str


class SageMakerRuntimeLike(Protocol):
    def invoke_endpoint(
        self,
        *,
        EndpointName: str,
        Body: Union[bytes, bytearray, IO[bytes]],
        ContentType: Optional[str] = None,
        Accept: Optional[str] = None,
        CustomAttributes: Optional[str] = None,
        TargetModel: Optional[str] = None,
        TargetVariant: Optional[str] = None,
        TargetContainerHostname: Optional[str] = None,
        InferenceId: Optional[str] = None,
        EnableExplanations: Optional[str] = None,
        InferenceComponentName: Optional[str] = None,
    ) -> InvokeEndpointOutput: ...



"""
A Python library for interacting with SageMaker deployments of the Isaacus API.
"""

from .types import IsaacusSageMakerRuntimeEndpoint
from .runtime_client import IsaacusSageMakerRuntimeHTTPClient, AsyncIsaacusSageMakerRuntimeHTTPClient



from isaacus_sagemaker._router import Router
from isaacus_sagemaker.types import IsaacusSageMakerRuntimeEndpoint


def test_router_requires_endpoints():
    try:
        Router([])
        assert False, "Expected assertion when no endpoints provided"
    except AssertionError:
        pass


def test_router_round_robin_with_model_and_default_mix():
    default = IsaacusSageMakerRuntimeEndpoint(name="default", region="us-east-1")
    # model-specific endpoint for "m1"
    m1a = IsaacusSageMakerRuntimeEndpoint(name="m1-a", region="us-east-1", models=["m1"])

    r = Router([default, m1a])

    # For model "m1" we should alternate between [m1a, default, m1a, default, ...]
    picks = [r.pick("m1").name for _ in range(6)]
    assert picks[:4] == ["m1-a", "default", "m1-a", "default"]

    # For some other model -> only defaults (here: a single one)
    assert r.pick("mX").name == "default"

    # When model is None -> from the defaults RR
    assert r.pick(None).name == "default"


def test_router_validates_endpoint_types_and_names():
    try:
        Router([object()])  # type: ignore[arg-type]
        assert False, "Expected assertion for wrong endpoint types"
    except AssertionError:
        pass

    try:
        Router([IsaacusSageMakerRuntimeEndpoint(name="", region="us-east-1")])
        assert False, "Expected assertion for empty endpoint name"
    except AssertionError:
        pass


import os
import json
import httpx
import msgspec
import pytest

from isaacus_sagemaker.runtime_client import (
    _extract_body,
    _get_model,
    _strip_unproxiable_headers,
    _make_payload,
    _translate_response,
    _build_request,
    IsaacusSageMakerRuntimeHTTPClient,
    AsyncIsaacusSageMakerRuntimeHTTPClient,
)
from isaacus_sagemaker.types import IsaacusSageMakerRuntimeEndpoint


def test_extract_body_accepts_json_and_rejects_others():
    req_ok = httpx.Request(
        method="POST",
        url="https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json"},
        content=b'{"hello":"world"}',
    )
    assert _extract_body(req_ok) == {"hello": "world"}

    # Non-JSON content types are not supported and should raise
    req_bad = httpx.Request(
        method="POST",
        url="https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "text/plain"},
        content=b"nope",
    )
    with pytest.raises(ValueError):
        _extract_body(req_bad)


def test_get_model_from_body():
    assert _get_model({"model": "kanon-2-embedder"}) == "kanon-2-embedder"
    assert _get_model({"not_model": "x"}) is None
    assert _get_model(None) is None  # type: ignore[arg-type]


def test_strip_unproxiable_headers():
    headers = {
        "Authorization": "Bearer abc",
        "Accept": "application/json",
        "Host": "api.isaacus.com",
        "X-Custom": "1",
        "Content-Length": "123",
    }
    out = _strip_unproxiable_headers(headers)
    assert "Authorization" not in out and "Host" not in out and "Content-Length" not in out
    assert out["Accept"] == "application/json"
    assert out["X-Custom"] == "1"


def test_make_payload_is_msgspec_json():
    req = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json", "accept": "application/json"},
        content=b'{"model":"m","x":1}',
    )
    body = _extract_body(req)
    payload = _make_payload(req, body)
    decoded = msgspec.json.decode(payload)
    assert decoded["path"] == "/v1/embeddings"
    assert decoded["method"] == "POST"
    assert decoded["data"] == {"model": "m", "x": 1}
    # headers should not contain authorization/host/etc
    hdrs = {k.lower(): v for k, v in decoded["headers"].items()}
    assert "authorization" not in hdrs and "host" not in hdrs
    assert hdrs["accept"] == "application/json"


def test_build_request_signs_and_targets_sagemaker(fake_aws, endpoints):
    client = IsaacusSageMakerRuntimeHTTPClient(endpoints=endpoints, region="ap-southeast-2")
    # ensure env var is injected when absent
    os.environ.pop("ISAACUS_API_KEY", None)
    client2 = IsaacusSageMakerRuntimeHTTPClient(endpoints=endpoints, region="ap-southeast-2")
    assert os.environ.get("ISAACUS_API_KEY") == "ISAACUS_SAGEMAKER_DOES_NOT_NEED_AN_API_KEY"

    # Build an original request
    req = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json", "accept": "application/json", "authorization": "Bearer abc"},
        content=b'{"model":"kanon-2-embedder","texts":["a"]}',
    )
    proxied = _build_request(client, req)

    # Target SageMaker invoke endpoint
    assert proxied.url.host.startswith("runtime.sagemaker.") and proxied.url.path.endswith("/invocations")
    assert proxied.headers.get("Authorization", "").startswith("AWS4-HMAC-SHA256 Dummy")
    assert proxied.headers["Content-Type"] == "application/json"
    assert proxied.headers["Accept"] == "application/json"

    # Payload should be the invocation struct JSON
    decoded = msgspec.json.decode(proxied.content)
    assert decoded["path"] == "/v1/embeddings"
    assert decoded["data"]["texts"] == ["a"]
    # ensure unproxiable headers were stripped from the inner payload
    inner_hdrs = {k.lower(): v for k, v in decoded["headers"].items()}
    assert "authorization" not in inner_hdrs


def test_build_request_raises_when_no_endpoint_for_model(fake_aws):
    client = IsaacusSageMakerRuntimeHTTPClient(
        endpoints=[IsaacusSageMakerRuntimeEndpoint(name="only-default", region="us-east-1", models=None)]
    )
    # Ask for a model *explicitly* that is not known and no default endpoints exist for it
    # (we simulate this by constructing a router with only a model-specific endpoint)
    client_only_model = IsaacusSageMakerRuntimeHTTPClient(
        endpoints=[IsaacusSageMakerRuntimeEndpoint(name="model-only", region="us-east-1", models=["m1"])]
    )
    req = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json"},
        content=b'{"model":"mX"}',
    )
    with pytest.raises(RuntimeError):
        _build_request(client_only_model, req)

    # sanity: with a default endpoint it should not raise
    _build_request(client, req)


def test_translate_response_passes_through_success_and_maps_wrapped_error():
    # 200 OK passthrough
    ok_resp = httpx.Response(200, json={"ok": True}, request=httpx.Request("GET", "https://x"))
    assert _translate_response(ok_resp) is ok_resp

    # Wrapped SageMaker error -> unwrap to original
    wrapped = httpx.Response(
        502,
        headers={"content-type": "application/json"},
        json={
            "OriginalStatusCode": 404,
            "OriginalMessage": json.dumps({"error": "Not Found"}),
        },
        request=httpx.Request("POST", "https://x"),
    )
    out = _translate_response(wrapped)
    assert out.status_code == 404
    assert out.headers["content-type"] == "application/json"
    assert out.json() == {"error": "Not Found"}


def test_sync_client_send_builds_and_translates(monkeypatch, fake_aws, endpoints):
    # Arrange: capture the request the base class would "send"
    sent = {}

    def fake_super_send(self, req, *args, **kwargs):
        # record what was sent
        sent["req"] = req
        # respond with a wrapped error to exercise translation
        return httpx.Response(
            500,
            headers={"content-type": "application/json"},
            json={
                "OriginalStatusCode": 400,
                "OriginalMessage": json.dumps({"error": "bad request"}),
            },
            request=req,
        )

    # Patch the base httpx.Client.send
    monkeypatch.setattr(httpx.Client, "send", fake_super_send, raising=True)

    client = IsaacusSageMakerRuntimeHTTPClient(endpoints=endpoints, region="ap-southeast-2")

    original = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json"},
        content=b'{"model":"kanon-2-embedder","texts":["t"]}',
    )
    resp = client.send(original)
    assert resp.status_code == 400
    assert resp.json() == {"error": "bad request"}

    # Ensure we actually proxied to SageMaker
    assert sent["req"].url.host.startswith("runtime.sagemaker.")


@pytest.mark.asyncio
async def test_async_client_skips_examples_host(monkeypatch, fake_aws, endpoints):
    # If targeting examples.isaacus.com we should *not* proxy or sign
    async def fake_async_send(self, req, *args, **kwargs):
        return httpx.Response(200, json={"ok": True}, request=req)

    # Guard: if _build_request is accidentally called, raise
    def boom(*args, **kwargs):  # pragma: no cover - ensures we don't proxy
        raise AssertionError("Should not have proxied requests to examples.isaacus.com")

    monkeypatch.setattr(httpx.AsyncClient, "send", fake_async_send, raising=True)

    # Also guard against accidental proxy
    import isaacus_sagemaker.runtime_client as rc
    monkeypatch.setattr(rc, "_build_request", boom, raising=True)

    client = AsyncIsaacusSageMakerRuntimeHTTPClient(endpoints=endpoints, region="ap-southeast-2")

    original = httpx.Request(
        "GET",
        "https://examples.isaacus.com/v1/echo",
    )
    resp = await client.send(original)
    assert resp.status_code == 200
    assert resp.json() == {"ok": True}


def test_get_creds_cached_by_profile(fake_aws, endpoints):
    import isaacus_sagemaker.runtime_client as rc

    # Construct a client with a *default* profile and one endpoint overriding it
    client = IsaacusSageMakerRuntimeHTTPClient(endpoints=endpoints, profile="dev", region="ap-southeast-2")

    # Two different requests should only construct a single boto3.Session per distinct profile key
    req1 = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json"},
        content=b'{"model":"kanon-2-embedder"}',
    )
    req2 = httpx.Request(
        "POST",
        "https://api.isaacus.com/v1/embeddings",
        headers={"content-type": "application/json"},
        content=b'{"model":"kanon-2-embedder"}',
    )
    _ = _build_request(client, req1)
    _ = _build_request(client, req2)

    # Only one session constructed for the same profile
    assert fake_aws["Boto3Session"].constructed == 1
